import scipy.sparse

from .dx import jump_normed, normed


def dx(dim, grid, bc, direction):
    deriv = normed(grid.n[dim], grid.N[dim], grid.h()[dim], bc, direction)
    if grid.ndim == 1:
        return deriv
    for d in range(0, grid.ndim):
        if d != dim:
            identity = scipy.sparse.identity(grid.n[d] * grid.N[d])
            if d < dim:
                deriv = scipy.sparse.kron(identity, deriv)
            else:
                deriv = scipy.sparse.kron(deriv, identity)
    return deriv


def jump(dim, grid, bc):
    deriv = jump_normed(grid.n[dim], grid.N[dim], grid.h()[dim], bc)
    if grid.ndim == 1:
        return deriv
    for d in range(0, grid.ndim):
        if d != dim:
            identity = scipy.sparse.identity(grid.n[d] * grid.N[d])
            if d < dim:
                deriv = scipy.sparse.kron(identity, deriv)
            else:
                deriv = scipy.sparse.kron(deriv, identity)
    return deriv



import numpy as np
import scipy.linalg
import scipy.sparse.linalg

from pyfeltor import dg


def test_elliptic():
    amp = 0.9
    def pol(y, x):
        return 1 + amp * np.sin(x) * np.sin(y)
    def rhs(y, x):
        return 2.0 * np.sin(x) * np.sin(y) * (amp * np.sin(x) * np.sin(y) + 1)\
            - amp * np.sin(x) * np.sin(x) * np.cos(y) * np.cos(y)\
            - amp * np.cos(x) * np.cos(x) * np.sin(y) * np.sin(y)
    def sol(y,x):
        return np.sin(x) * np.sin(y)
    lx, ly = np.pi, 2 * np.pi
    bcx, bcy = dg.bc.DIR, dg.bc.PER
    n, Nx, Ny = 3, 64, 64
    jfactor = 1
    grid = dg.Grid([0, 0], [ly, lx], [n, n], [Ny, Nx])
    w2d = dg.create.weights(grid)
    x = np.zeros(grid.size())
    b = dg.evaluate(rhs, grid)
    chi = dg.evaluate(pol, grid)
    solution = dg.evaluate(sol, grid)
    print("Forward elliptic ")
    pol_forward = dg.create.elliptic(
        grid,
        [bcy, bcx],
        [dg.direction.forward, dg.direction.forward],
        sigma=chi,
        jumpfactor=jfactor,
    )
    error = pol_forward.dot(solution) - b
    print(
        "Direct application of laplace to solution is",
        np.sqrt(np.sum(w2d * error ** 2)),
    )
    assert np.isclose(np.sqrt(np.sum(w2d * error ** 2)), 0.07475654581481737)
    x = scipy.sparse.linalg.spsolve(pol_forward, b)

    print("Distance to true solution is ", np.sqrt(np.sum(w2d * (x - solution) ** 2)))


def test_elliptic1d():
    print("TEST 1d ELLIPTIC")
    grid1d = dg.Grid([0], [np.pi], [3], [4])
    w1d = dg.create.weights(grid1d)
    x = np.zeros(grid1d.size())
    b = dg.evaluate(lambda x: np.sin(x), grid1d)
    chi = np.ones(grid1d.size())
    solution1d = dg.evaluate(lambda x: np.sin(x), grid1d)
    pol_backward = dg.create.elliptic(
        grid1d, [dg.bc.DIR], [dg.direction.backward], sigma=chi, jumpfactor=1
    )
    x = pol_backward.dot(solution1d) - b
    assert np.isclose(np.sqrt(np.sum(w1d * x ** 2)), 0.28502509566157097)



import json

import numpy as np

from pyfeltor import dg


def test_polynomial():
    c = np.array([1, 2, 3, 4])
    params = {"R_0": 400, "inverseaspectratio": 20, "elongation": 1, "triangularity": 1,
              "PP": 1, "PI": 1, "description": "standardX", "M": 2, "N": 2,
              "c": c.tolist()}
    pp = dg.geo.polynomial.Parameters(params)
    psip = dg.geo.polynomial.Psip(pp)
    grid = dg.Grid(x0=(pp.R_0 - pp.a, -pp.a), x1=(pp.R_0 + pp.a, pp.a),
        n=(3, 3), N=(24, 24))
    psi = dg.evaluate(psip, grid)
    R = dg.evaluate(lambda R, Z: R, grid)
    Z = dg.evaluate(lambda R, Z: Z, grid)
    psi_ = psip(R, Z)
    print(psi)
    print(psi_)


def test_make_field(datadir):
    with open(datadir / "geometry_params_Xpoint.json") as f:
        magparams = json.load(f)
    mag = dg.geo.createMagneticField(magparams)
    a = mag.params().a()
    R0 = mag.R0()
    grid = dg.Grid(x0=(R0 - a, -a), x1=(R0 + a, +a), n=(3, 3), N=(24, 24))
    psi = dg.evaluate(mag.psip(), grid)
    BR = dg.evaluate(dg.geo.BFieldR(mag), grid)
    print(mag.R0())
    print(mag.params().a())
    print(mag.params().elongation())
    print(mag.params().getDescription())
    print(mag.params().getEquilibrium())
    print(psi, BR)


def test_q_profile(datadir):
    with open(datadir / "enrx_tcv.json") as f:
        magparams = json.load(f)
    mag = dg.geo.createMagneticField(magparams)
    qfunctor = dg.geo.SafetyFactor(mag)
    RO, ZO = mag.R0(), 0
    (point, RO, ZO) = dg.geo.findOpoint(mag.get_psip(), RO, ZO)
    print("O-point found at ", RO, ZO)
    psipO = mag.psip()(RO, ZO)
    psi_values = np.linspace(psipO, 0, 20, endpoint=False)
    print(qfunctor(psi_values))


def test_sheath(datadir):
    with open(datadir / "enrx_tcv.json") as f:
        magparams = json.load(f)
    mag = dg.geo.createMagneticField(magparams)
    wall = {"type": "sol_pfr", "alpha": [0.05, 0.05], "boundary": [1.09, 0.97]}
    sheath = {"boundary": 3 / 32, "alpha": 2 / 32}
    wall_f = dg.geo.CylindricalFunctor()
    transition_f = dg.geo.CylindricalFunctor()
    sheath_f = dg.geo.CylindricalFunctor()
    dg.geo.createModifiedField(magparams, wall, wall_f, transition_f)
    a = mag.params().a()
    R0 = mag.R0()
    print(R0 - a, -a, R0 + a, +a)
    grid = dg.Grid(x0=(R0 - a, -a), x1=(R0 + a, +a), n=(3, 3), N=(24, 24))
    dg.geo.createSheathRegion(sheath, mag, wall_f, grid, sheath_f)

    point, RO, ZO = dg.geo.findOpoint(mag.get_psip(), R0, 0)
    assert (point == 1)
    assert (abs(RO - 0.9022324795511738) < 1e-10)
    assert (abs(ZO - (-0.005684720497897021)) < 1e-10)

    closed = dg.geo.mod.ClosedFieldlineRegion(mag)
    assert (closed(0, R0))
    opened = dg.geo.mod.ClosedFieldlineRegion(mag, False)
    assert (not opened(0, R0))
    sol = dg.geo.mod.SOLRegion(mag, wall_f)
    assert (not sol(0, R0))

    cutter = dg.geo.ZCutter(10)
    assert (cutter(1, 11))
    assert (not cutter(1, 8))



import numpy as np

from ..grid import Grid


def weights(grid):
    """Create dG weights on given grid"""
    weights = np.array([1])
    for dim in range(0, grid.ndim):
        (x, w) = np.polynomial.legendre.leggauss(grid.n[dim])
        wdim = np.tile(w, grid.N[dim]) * grid.h()[dim] / 2.0
        weights = np.kron(weights, wdim)
        # weights = np.reshape(weights, grid.shape[0 : dim + 1])
    return weights


def abscissas(grid, dimension=0):
    """Create 1d dG abscissas on given grid for given dimension"""
    dim = dimension
    (x, w) = np.polynomial.legendre.leggauss(grid.n[dim])
    abscissas = np.zeros(grid.shape[dim])
    h = grid.h()[dim]
    x0 = grid.x0[dim]
    for i in range(0, grid.N[dim]):
        for k in range(0, grid.n[dim]):
            abscissas[i * grid.n[dim] + k] = i * h + x0 + h / 2.0 * (1.0 + x[k])
    return abscissas


def _get_x0_x1_n_N(x):
    for n in range(1, 21):
        if len(x) % n != 0:
            continue
        N = len(x) / n
        (xx, w) = np.polynomial.legendre.leggauss(n)
        h = x[n] - x[0]
        x0 = x[0] - h / 2.0 * (1.0 + xx[0])
        first_absc = np.zeros(2 * n)
        last_value = x0 + h * N - h / 2.0 * (1.0 + xx[0])
        for i in range(0, 2):
            for k in range(0, n):
                first_absc[i * n + k] = i * h + x0 + h / 2.0 * (1.0 + xx[k])
        if abs(last_value - x[-1]) < 1e-10 * abs(x[-1]) + 1e-10:
            return (x0, x0 + h * N, int(n), int(N))
    raise Exception("Could not determine grid")
    return (0, 1, 1, 1)


def grid_from_abscissas(xs):
    """ This function reverse engineers a Grid that corresponds to the given
    abscissas. This is useful for example to get a Grid that corresponds to a
    given simulation output

    xs(list of 1d arrays): the length of the list is ndim. Last entry
        is the x-dimension. Each entry is a 1d array corresponding to the abscissas
        in that dimension. At least two cells need to be present.
    Return: Grid whose abscissas correspond to xs
    """
    ndim = len(xs)
    x0 = np.zeros(ndim)
    x1 = np.zeros(ndim)
    n = np.arange(ndim)
    N = np.arange(ndim)
    for dim in range(0, ndim):
        x0[dim], x1[dim], n[dim], N[dim] = _get_x0_x1_n_N(xs[dim])
    return Grid(x0, x1, n, N)



# docstring displayed by help(pyfeltor)
""" The python version of the dg library
"""

# Import into their own dg.geo.xxx namespace
from . import circular as circular
from . import guenter as guenter
from . import mod as mod
from . import polynomial as polynomial
from . import solovev as solovev
from . import toroidal as toroidal

# Import into the dg.geo namespace
from .flux import *
from .geometries import *
from .utility import *



from enum import Enum


class bc(Enum):
    PER = (0,)
    DIR = (1,)
    DIR_NEU = (2,)
    NEU_DIR = (3,)
    NEU = 4


def inverse_bc(bound):
    """return the inverse boundary condition"""
    if bound == bc.DIR:
        return bc.NEU
    if bound == bc.NEU:
        return bc.DIR
    if bound == bc.DIR_NEU:
        return bc.NEU_DIR
    if bound == bc.NEU_DIR:
        return bc.DIR_NEU
    return bc.PER


class direction(Enum):
    forward = (0,)
    backward = (1,)
    centered = 2


def inverse_dir(direct):
    """return the inverse direction"""
    if direct == direction.forward:
        return direction.backward
    if direct == direction.backward:
        return direction.forward
    return direction.centered



import numpy as np

from . import create
from .create import operators as ops
from .enums import direction


def evaluate(function, grid):
    """Evaluate a function on the grid points of the given grid

    function: has to take numpy arrays as arguments, f(x), f(y,x), f(z,y,x)
    grid: instance of dg.Grid
    return: flat np.array with x the fastest varying dimension. Can be reshaped
    with reshape(grid.shape)
    """
    ndim = grid.ndim
    xs = [create.abscissas(grid, dim) for dim in range(0, ndim)]

    if ndim == 1:
        return np.array([function(x) for x in xs[0]])
    if ndim == 2:
        return np.array([function(y, x) for y in xs[0] for x in xs[1]])
    if ndim == 3:
        return np.array(
            [function(z, y, x) for z in xs[0] for y in xs[1] for x in xs[2]]
        )
    if ndim == 4:
        return np.array(
            [
                function(xx, z, y, x)
                for xx in xs[0]
                for z in xs[1]
                for y in xs[2]
                for x in xs[3]
            ]
        )
    if ndim > 4:
        raise Exception("dg.evaluate is not implemented for ndim > 4")


def integrate(to_integrate, grid, direction=direction.forward):
    # python passes parameters "by assignment"
    if grid.ndim > 1:
        raise Exception("dg.integrate not implemented for ndim > 1")
    h = grid.h()[0]
    n = grid.n[0]
    to_in = to_integrate
    if direction == direction.backward:  # reverse input vector
        # https://stackoverflow.com/questions/6771428/most-efficient-way-to-reverse-a-numpy-array
        to_in = to_integrate[::-1]  # create a reversed view

    forward = ops.forward(n)
    backward = ops.backward(n)
    ninj = ops.ninj(n)
    t = ops.pipj_inv(n) * h / 2.0
    ninj = backward @ t @ ninj @ forward
    constant = 0.0

    out = np.zeros(grid.size())
    for i in range(0, grid.N[0]):
        for k in range(0, grid.n[0]):
            for ll in range(0, grid.n[0]):
                out[i * n + k] += ninj[k, ll] * to_in[i * n + ll]
            out[i * n + k] += constant
        for k in range(0, grid.n[0]):
            constant += h * forward[0, k] * to_in[i * n + k]
    if direction == direction.backward:  # reverse output
        out = -out[::-1]
    return out



# docstring displayed by help(pyfeltor)
""" The python version of the dg library
"""
from . import create as create
from . import geo as geo
from .enums import bc as bc
from .enums import direction as direction
from .enums import inverse_bc as inverse_bc
from .enums import inverse_dir as inverse_dir
from .evaluation import evaluate as evaluate
from .evaluation import integrate as integrate
from .grid import Grid as Grid



import itertools

import numpy as np
import scipy.sparse

from ..enums import bc
from ..evaluation import evaluate
from . import operators as ops
from .weights import weights


def shift(grid, x, bcs):
    """ Shift a point into the grid according to boundary conditions
        (used internally by interpolation)
    """
    negative = False
    for i in range(0, grid.ndim):
        if bcs[i] == bc.PER:
            N = np.floor(
                (x[i] - grid.x0[i]) / (grid.x1[i] - grid.x0[i])
            )  # ... -2[ -1[ 0[ 1[ 2[ ...
            x[i] = x[i] - N * (grid.x1[i] - grid.x0[i])  # shift
        # mirror along boundary as often as necessary
        while (x[i] < grid.x0[i]) or (x[i] > grid.x1[i]):
            if x[i] < grid.x0[i]:
                x[i] = 2.0 * grid.x0[i] - x[i]
                # every mirror swaps the sign if Dirichlet
                if (bcs[i] == bc.DIR) or (bcs[i] == bc.DIR_NEU):
                    negative = not negative  # swap sign
            if x[i] > grid.x1[i]:
                x[i] = 2.0 * grid.x1[i] - x[i]
                if (bcs[i] == bc.DIR) or (bcs[i] == bc.NEU_DIR):
                    # notice the different boundary NEU_DIR to the above DIR_NEU !
                    negative = not negative  # swap sign
    return negative, x


def interpolation(xs, grid, bcs):
    """ interpolation matrix that evaluate points on given grid

        order is xs = [... z, y, x]
    """

    rows = []
    cols = []
    vals = []
    forward = [ops.forward(grid.n[dim]) for dim in range(0, grid.ndim)]

    for i in range(0, len(xs[0])):
        X = np.array([xs[k][i] for k in range(0, grid.ndim)])
        negative, X = shift(grid, X, bcs)
        xnn = (X - grid.x0) / grid.h()
        nn = np.floor(xnn)
        xn = 2.0 * xnn - (2 * nn + 1)
        for k in range(0, grid.ndim):
            if nn[k] == grid.N[k]:
                nn[k] -= 1
                xn[k] = 1

        # evaluate Legendre polynomials at (xn)...
        px = [
            (
                np.polynomial.legendre.legvander(xn[k], grid.n[k] - 1) @ forward[k]
            ).ravel()
            for k in range(0, grid.ndim)
        ]

        # generate the list of index permutations
        tuples = []
        if grid.ndim == 1:
            tuples = [(a,) for a in range(0, grid.n[0])]
        elif grid.ndim == 2:
            tuples = itertools.product(range(0, grid.n[0]), range(0, grid.n[1]))
        elif grid.ndim == 3:
            tuples = itertools.product(
                range(0, grid.n[0]), range(0, grid.n[1]), range(0, grid.n[2])
            )
        else:
            raise Exception("interpolation not implemented for ndim > 3")
        for it in tuples:
            II = 0  # the index to push back
            VV = 1  # the value to push back
            for kk in range(0, grid.ndim):
                II = (II * grid.N[kk] + nn[kk]) * grid.n[kk] + it[kk]
                VV = VV * px[kk][it[kk]]
            # print( i, I, V)
            rows.append(i)
            cols.append(round(II))
            if not negative:
                vals.append(VV)
            else:
                vals.append(-VV)
    # sort
    rows, cols, vals = zip(*sorted(zip(rows, cols, vals, strict=False)), strict=False)
    return scipy.sparse.coo_matrix((vals, (rows, cols)), shape=(len(xs[0]),
        grid.size()))


def projection(grid_new, grid_old):
    """ Create a projection between two grids

    This matrix can be applied to vectors defined on the old (fine) grid to obtain
    its values projected on the new (coarse) grid. (Projection means that the
    projection integrals over the base polynomials are computed).
    The projection matrix is the adjoint of the interpolation matrix
    params:
    g_new : The new (coarse) grid
    g_old : The old (fine) grid
    """
    ndim = grid_old.ndim
    if grid_new.ndim != ndim:
        raise Exception("Cannot project between grids with different dimensions")
    for i in range(0, ndim):
        if grid_old.N[i] % grid_new.N[i] != 0:
            print(f"WARNING you project between incompatible grids!! old N: \
{grid_old.N[i]} new N {grid_new.N[i]}")
        if grid_old.n[i] < grid_new.n[i]:
            print(f"WARNING you project between incompatible grids!! old n: \
{grid_old.n[i]} new n {grid_new.n[i]}")
    wf = scipy.sparse.diags(weights(grid_old))
    points = []
    bcs = [bc.PER for i in range(0, ndim)]
    if ndim == 1:
        points.append(evaluate(lambda x: x, grid_old))
    elif ndim == 2:
        points.append(evaluate(lambda y, x: y, grid_old))
        points.append(evaluate(lambda y, x: x, grid_old))
    elif ndim == 3:
        points.append(evaluate(lambda z, y, x: z, grid_old))
        points.append(evaluate(lambda z, y, x: y, grid_old))
        points.append(evaluate(lambda z, y, x: x, grid_old))
    else:
        raise Exception("Projection not implemented for ndim > 3")
    A = interpolation(points, grid_new, bcs)
    vc = scipy.sparse.diags(1. / weights(grid_new))
    return vc @ A.transpose() @ wf



import numpy as np

from pyfeltor import dg

# Run with pytest-3 -s . to see stdout output


def test_grid():
    print("TEST GRID")
    g1d = dg.Grid(x0=1, x1=2, n=3, N=24)
    print(g1d.x0, g1d.x1, g1d.n, g1d.N)
    assert g1d.ndim == 1
    g2d = dg.Grid(x0=(0, 0), x1=(1, 1), n=(3, 3), N=(24, 24))
    print(g2d.x0, g2d.x1, g2d.n, g2d.N)
    assert g2d.ndim == 2
    assert g2d.size() == 5184


def test_creation():
    print("TEST CREATION")
    grid = dg.Grid(x0=(0, 0), x1=(1, 1), n=(3, 3), N=(24, 24))
    weights = dg.create.weights(grid)
    print(weights)


def test_evaluation():
    print("TEST EVALUATION")
    grid = dg.Grid(x0=(0, 0), x1=(np.pi, np.pi), n=(3, 3), N=(24, 24))
    weights = dg.create.weights(grid)
    sine = dg.evaluate(lambda y, x: np.sin(y) * np.sin(x), grid)
    print(f"Integral is {np.sum(sine * weights)}")


def test_grid_from_abscissas():
    print("TEST REVERSE GRID ENGINEER")
    grid = dg.Grid(x0=(-10, 0, 20, -10, 0, 20), x1=(2 * np.pi, np.pi, 10 * np.pi,
        2 * np.pi, np.pi, 10 * np.pi), n=(1, 2, 3, 4, 5, 6), N=(20, 24, 12, 13, 15, 16))
    xs = [dg.create.abscissas(grid, i) for i in range(0, grid.ndim)]
    test_grid = dg.create.grid_from_abscissas(xs)

    print(grid.x0, grid.x1, grid.n, grid.N)
    print(test_grid.x0, test_grid.x1, test_grid.n, test_grid.N)
    assert np.allclose(grid.x0, test_grid.x0)
    assert np.allclose(grid.x1, test_grid.x1)
    assert np.allclose(grid.n, test_grid.n)
    assert np.allclose(grid.N, test_grid.N)



import scipy.sparse

from ..enums import bc
from . import operators as ops


def symm(n, N, h, bcx):
    ll = ops.lilj(n)
    rr = ops.rirj(n)
    lr = ops.lirj(n)
    rl = ops.rilj(n)
    d = ops.pidxpj(n)
    t = ops.pipj_inv(n)
    t *= 2.0 / h

    a = 1.0 / 2.0 * t @ (d - d.transpose())  # @ is matrix multiplication
    # bcx = PER
    a_bound_right = a.copy()
    a_bound_left = a.copy()
    # left boundary
    if (bcx == bc.DIR) or (bcx == bc.DIR_NEU):
        a_bound_left += 0.5 * (t @ ll)
    elif (bcx == bc.NEU) or (bcx == bc.NEU_DIR):
        a_bound_left -= 0.5 * (t @ ll)
    # right boundary
    if (bcx == bc.DIR) or (bcx == bc.NEU_DIR):
        a_bound_right -= 0.5 * (t @ rr)
    elif (bcx == bc.NEU) or (bcx == bc.DIR_NEU):
        a_bound_right += 0.5 * (t @ rr)
    b = t @ (1.0 / 2.0 * rl)
    bp = t @ (-1.0 / 2.0 * lr)  # pitfall: T*-m^T is NOT -(T*m)^T
    # transform to XSPACE
    backward = ops.backward(n)
    forward = ops.forward(n)
    a = (backward @ a) @ forward
    b = (backward @ b) @ forward
    bp = (backward @ bp) @ forward
    a_bound_left = (backward @ a_bound_left) @ forward
    a_bound_right = (backward @ a_bound_right) @ forward
    # assemble the matrix
    rows = []
    cols = []
    vals = []
    if bcx != bc.PER:
        for i in range(0, n):
            for j in range(0, n):
                rows.append(0 * n + i)
                cols.append((0 * n + j) % (n * N))
                vals.append(a_bound_left[i, j])

                rows.append(0 * n + i)
                cols.append(((0 + 1) * n + j) % (n * N))
                vals.append(b[i, j])

        for k in range(1, N - 1):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append(((k - 1) * n + j) % (n * N))
                    vals.append(bp[i, j])

                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

                    rows.append(k * n + i)
                    cols.append(((k + 1) * n + j) % (n * N))
                    vals.append(b[i, j])

        for i in range(0, n):
            for j in range(0, n):
                rows.append((N - 1) * n + i)
                cols.append((((N - 1) - 1) * n + j) % (n * N))
                vals.append(bp[i, j])

                rows.append((N - 1) * n + i)
                cols.append(((N - 1) * n + j) % (n * N))
                vals.append(a_bound_right[i, j])

    else:  # periodic
        for k in range(0, N):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append(((k - 1) * n + j) % (n * N))
                    vals.append(bp[i, j])

                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

                    rows.append(k * n + i)
                    cols.append(((k + 1) * n + j) % (n * N))
                    vals.append(b[i, j])

    # sort
    rows, cols, vals = zip(*sorted(zip(rows, cols, vals, strict=False)), strict=False)
    return scipy.sparse.coo_matrix((vals, (rows, cols)))


def plus(n, N, h, bcx):
    ll = ops.lilj(n)
    #rr = ops.rirj(n)
    #lr = ops.lirj(n)
    rl = ops.rilj(n)
    d = ops.pidxpj(n)
    t = ops.pipj_inv(n)
    t *= 2.0 / h

    a = t @ (-ll - d.transpose())
    # bcx = PER
    a_bound_left = a.copy()  # PER, NEU, and NEU_DIR
    a_bound_right = a.copy()  # PER, DIR, and NEU_DIR
    if (bcx == bc.DIR) or (bcx == bc.DIR_NEU):
        a_bound_left = t @ (-d.transpose())
    if (bcx == bc.NEU) or (bcx == bc.DIR_NEU):
        a_bound_right = t @ d
    b = t @ rl
    # transform to XSPACE
    backward = ops.backward(n)
    forward = ops.forward(n)
    a = (backward @ a) @ forward
    b = (backward @ b) @ forward
    a_bound_left = (backward @ a_bound_left) @ forward
    a_bound_right = (backward @ a_bound_right) @ forward
    # assemble the matrix
    rows = []
    cols = []
    vals = []
    if bcx != bc.PER:
        for i in range(0, n):
            for j in range(0, n):
                rows.append(0 * n + i)
                cols.append((0 * n + j) % (n * N))
                vals.append(a_bound_left[i, j])

                rows.append(0 * n + i)
                cols.append(((0 + 1) * n + j) % (n * N))
                vals.append(b[i, j])

        for k in range(1, N - 1):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

                    rows.append(k * n + i)
                    cols.append(((k + 1) * n + j) % (n * N))
                    vals.append(b[i, j])

        for i in range(0, n):
            for j in range(0, n):
                rows.append((N - 1) * n + i)
                cols.append(((N - 1) * n + j) % (n * N))
                vals.append(a_bound_right[i, j])

    else:  # periodic
        for k in range(0, N):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

                    rows.append(k * n + i)
                    cols.append(((k + 1) * n + j) % (n * N))
                    vals.append(b[i, j])

    # sort
    rows, cols, vals = zip(*sorted(zip(rows, cols, vals, strict=False)), strict=False)
    return scipy.sparse.coo_matrix((vals, (rows, cols)))


def minus(n, N, h, bcx):
    ll = ops.lilj(n)
    #rr = ops.rirj(n)
    lr = ops.lirj(n)
    #rl = ops.rilj(n)
    d = ops.pidxpj(n)
    t = ops.pipj_inv(n)
    t *= 2.0 / h

    a = t @ (ll + d)
    # bcx = PER
    a_bound_right = a.copy()  # PER, NEU and DIR_NEU
    a_bound_left = a.copy()  # PER, DIR and DIR_NEU
    if (bcx == bc.DIR) or (bcx == bc.NEU_DIR):
        a_bound_right = t @ (-d.transpose())
    if (bcx == bc.NEU) or (bcx == bc.NEU_DIR):
        a_bound_left = t @ d
    bp = -t @ lr
    # transform to XSPACE
    backward = ops.backward(n)
    forward = ops.forward(n)
    a = (backward @ a) @ forward
    bp = (backward @ bp) @ forward
    a_bound_left = (backward @ a_bound_left) @ forward
    a_bound_right = (backward @ a_bound_right) @ forward
    # assemble the matrix
    rows = []
    cols = []
    vals = []
    if bcx != bc.PER:
        for i in range(0, n):
            for j in range(0, n):
                rows.append(0 * n + i)
                cols.append((0 * n + j) % (n * N))
                vals.append(a_bound_left[i, j])

        for k in range(1, N - 1):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append(((k - 1) * n + j) % (n * N))
                    vals.append(bp[i, j])

                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

        for i in range(0, n):
            for j in range(0, n):
                rows.append((N - 1) * n + i)
                cols.append((((N - 1) - 1) * n + j) % (n * N))
                vals.append(bp[i, j])

                rows.append((N - 1) * n + i)
                cols.append(((N - 1) * n + j) % (n * N))
                vals.append(a_bound_right[i, j])

    else:  # periodic
        for k in range(0, N):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append(((k - 1) * n + j) % (n * N))
                    vals.append(bp[i, j])

                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

    # sort
    rows, cols, vals = zip(*sorted(zip(rows, cols, vals, strict=False)), strict=False)
    return scipy.sparse.coo_matrix((vals, (rows, cols)))


def jump_normed(n, N, h, bcx):
    ll = ops.lilj(n)
    rr = ops.rirj(n)
    lr = ops.lirj(n)
    rl = ops.rilj(n)

    t = ops.pipj_inv(n)
    t *= 2.0 / h
    a = t @ (ll + rr)
    a_bound_left = a.copy()  # DIR and PER
    if (bcx == bc.NEU) or (bcx == bc.NEU_DIR):
        a_bound_left = t @ rr
    a_bound_right = a.copy()  # DIR and PER
    if (bcx == bc.NEU) or (bcx == bc.DIR_NEU):
        a_bound_right = t @ ll
    b = -t @ rl
    bp = -t @ lr
    # transform to XSPACE
    backward = ops.backward(n)
    forward = ops.forward(n)
    a = (backward @ a) @ forward
    b = (backward @ b) @ forward
    bp = (backward @ bp) @ forward
    a_bound_left = (backward @ a_bound_left) @ forward
    a_bound_right = (backward @ a_bound_right) @ forward
    # assemble the matrix
    rows = []
    cols = []
    vals = []
    if bcx != bc.PER:
        for i in range(0, n):
            for j in range(0, n):
                rows.append(0 * n + i)
                cols.append((0 * n + j) % (n * N))
                vals.append(a_bound_left[i, j])

                rows.append(0 * n + i)
                cols.append(((0 + 1) * n + j) % (n * N))
                vals.append(b[i, j])

        for k in range(1, N - 1):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append(((k - 1) * n + j) % (n * N))
                    vals.append(bp[i, j])

                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

                    rows.append(k * n + i)
                    cols.append(((k + 1) * n + j) % (n * N))
                    vals.append(b[i, j])

        for i in range(0, n):
            for j in range(0, n):
                rows.append((N - 1) * n + i)
                cols.append((((N - 1) - 1) * n + j) % (n * N))
                vals.append(bp[i, j])

                rows.append((N - 1) * n + i)
                cols.append(((N - 1) * n + j) % (n * N))
                vals.append(a_bound_right[i, j])

    else:  # periodic
        for k in range(0, N):
            for i in range(0, n):
                for j in range(0, n):
                    rows.append(k * n + i)
                    cols.append(((k - 1) * n + j) % (n * N))
                    vals.append(bp[i, j])

                    rows.append(k * n + i)
                    cols.append((k * n + j) % (n * N))
                    vals.append(a[i, j])

                    rows.append(k * n + i)
                    cols.append(((k + 1) * n + j) % (n * N))
                    vals.append(b[i, j])

    # sort
    rows, cols, vals = zip(*sorted(zip(rows, cols, vals, strict=False)), strict=False)
    return scipy.sparse.coo_matrix((vals, (rows, cols)))


def normed(n, N, h, bcx, direction):
    if direction == direction.centered:
        return symm(n, N, h, bcx)
    elif direction == direction.forward:
        return plus(n, N, h, bcx)
    elif direction == direction.backward:
        return minus(n, N, h, bcx)



import numpy as np

from pyfeltor import dg

# Run with pytest-3 -s . to see stdout output


def zero(y, x):
    return 0


def sine(y, x):
    return np.sin(x) * np.sin(y)


def cosx(y, x):
    return np.cos(x) * np.sin(y)


def cosy(y, x):
    return np.cos(y) * np.sin(x)


def test_dx():
    n, N = 3, 20
    gPER = dg.Grid(0.1, 2 * np.pi + 0.1, n, N)
    gDIR = dg.Grid(0, np.pi, n, N)
    gNEU = dg.Grid(np.pi / 2.0, 3 * np.pi / 2.0, n, N)
    gDIR_NEU = dg.Grid(0, np.pi / 2.0, n, N)
    gNEU_DIR = dg.Grid(np.pi / 2.0, np.pi, n, N)
    grids = [gPER, gDIR, gNEU, gDIR_NEU, gNEU_DIR]
    bcxs = [dg.bc.PER, dg.bc.DIR, dg.bc.NEU, dg.bc.DIR_NEU, dg.bc.NEU_DIR]
    print("TEST NORMAL TOPOLOGY: YOU SHOULD SEE CONVERGENCE FOR ALL OUTPUTS!!!")
    print("COMPARE TO dx_t.cu")
    for (g, bcx) in zip(grids, bcxs, strict=False):
        print("Boundary condition ", bcx)
        hs = dg.create.dx(0, g, bcx, dg.direction.centered)
        hf = dg.create.dx(0, g, bcx, dg.direction.forward)
        hb = dg.create.dx(0, g, bcx, dg.direction.backward)
        js = dg.create.jump(0, g, bcx)
        func = dg.evaluate(lambda x: np.sin(x), g)
        error = func
        w1d = dg.create.weights(g)
        deri = dg.evaluate(lambda x: np.cos(x), g)
        null = dg.evaluate(lambda x: 0, g)
        error = deri - hs.dot(func)
        print(
            f"Distance to true solution (symmetric): {np.sqrt(np.sum(w1d * error**2))}"
        )
        error = deri - hf.dot(func)
        print(f"Distance to true solution (forward): {np.sqrt(np.sum(w1d * error**2))}")
        error = deri - hb.dot(func)
        print(
            f"Distance to true solution (backward): {np.sqrt(np.sum(w1d * error**2))}"
        )
        error = null - js.dot(func)
        print(
            f"Distance to true solution (jump     ): {np.sqrt(np.sum(w1d * error**2))}"
        )


def test_derivative():
    n, Nx, Ny = 3, 24, 28
    print(f"On Grid {n} x {Nx} x {Ny}")
    bcx, bcy = dg.bc.DIR, dg.bc.PER
    g2d = dg.Grid([0.1, 0], [2 * np.pi + 0.1, np.pi], [n, n], [Ny, Nx])
    w2d = dg.create.weights(g2d)

    dx2 = dg.create.dx(1, g2d, bcx, dg.direction.forward)
    dy2 = dg.create.dx(0, g2d, bcy, dg.direction.centered)
    jx2 = dg.create.jump(1, g2d, bcx)
    jy2 = dg.create.jump(0, g2d, bcy)
    m2 = [dx2, dy2, jx2, jy2]
    f2d = dg.evaluate(sine, g2d)
    dx2d = dg.evaluate(cosx, g2d)
    dy2d = dg.evaluate(cosy, g2d)
    null2 = dg.evaluate(zero, g2d)
    sol2 = [dx2d, dy2d, null2, null2]
    sol = [
        0.0010775034079703078,
        0.00027314872436790867,
        0.0021410107540649087,
        0.006471682818855125,
    ]

    print("TEST 2D: DX, DY, JX, JY")
    for i in range(0, 4):
        error = sol2[i].copy()
        error = -m2[i].dot(f2d) + error
        norm = np.sqrt(np.sum(w2d * error ** 2))
        print(f"Absolute error to true solution: {norm}")
        assert np.isclose(norm, sol[i])



import numpy as np

from pyfeltor import dg


def test_interpolation1d():
    print("1D INTERPOLATION TEST")
    passed = True
    n, Nx = 3, 9
    g1d = dg.Grid(-np.pi, 0, n, Nx)
    field = dg.evaluate(lambda x: x, g1d)
    x = [(i + 0.5) * g1d.h()[0] / g1d.n[0] for i in range(0, g1d.size())]

    interp = dg.create.interpolation([x], g1d, [dg.bc.DIR])
    x_inter = interp.dot(field)
    for i in range(g1d.size()):
        if x[i] - x_inter[i] > 1e-14:
            print(f"X NOT EQUAL {i}\t{x[i]}  \t{x_inter[i]}")
            passed = False
    assert passed


def test_interpolation2d():
    print("2D INTERPOLATION TEST")

    n, Nx, Ny = 3, 32, 32
    g = dg.Grid(x0=[-5 * np.pi, -np.pi], x1=[-4 * np.pi, 0], n=[n, n], N=[Ny, Nx])
    equi = dg.Grid(
        x0=[-5 * np.pi, -np.pi], x1=[-4 * np.pi, 0], n=[1, 1], N=[n * Ny, n * Nx]
    )
    y = dg.evaluate(lambda y, x: y, equi)
    x = dg.evaluate(lambda y, x: x, equi)

    interp = dg.create.interpolation([y, x], g, [dg.bc.DIR, dg.bc.DIR])
    vec = dg.evaluate(lambda y, x: np.sin(x) * np.sin(y), g)
    inter = interp.dot(vec)
    interE = dg.evaluate(lambda y, x: np.sin(x) * np.sin(y), equi)
    error = np.sum((inter - interE) ** 2) / np.sum(inter ** 2)
    print(f"Error is {np.sqrt(error)} (should be small)!")
    assert np.isclose(np.sqrt(error), 0, atol=1e-5)


def test_interpolation3d():
    print("3D INTERPOLATION TEST")
    n, Nx, Ny, Nz = 3, 9, 5, 4

    g = dg.Grid([-7, -5, -10], [-3, 5, 10], [1, n, n], [Nz, Ny, Nx])
    equi = dg.Grid([-7, -5, -10], [-3, 5, 10], [1, 1, 1], [1 * Nz, n * Ny, n * Nx])
    z = dg.evaluate(lambda z, y, x: z, equi)
    y = dg.evaluate(lambda z, y, x: y, equi)
    x = dg.evaluate(lambda z, y, x: x, equi)

    interp = dg.create.interpolation([z, y, x], g, [dg.bc.DIR, dg.bc.DIR, dg.bc.PER])
    vec = dg.evaluate(lambda z, y, x: np.sin(x) * np.sin(y) * np.sin(z), g)
    inter = interp.dot(vec)
    # These values are from interpolation_t.cu
    solution = np.array(
        [
            -0.0388107,
            0.108781,
            0.200279,
            0.188096,
            0.07971,
            -0.0697783,
            -0.189282,
            -0.20544,
            -0.115663,
            0.0414357,
        ]
    )
    print("Pointwise error ", inter[0:10] - solution)
    assert np.allclose(inter[0:10], solution)


def test_projection():
    print("2D PROJECTION TEST")

    n, Nx, Ny = 3, 8, 8
    g = dg.Grid(x0=[-5 * np.pi, -np.pi], x1=[-4 * np.pi, 0], n=[n, n], N=[Ny, Nx])
    g_fine = dg.Grid(x0=g.x0, x1=g.x1, n=g.n, N=[n * Ny, n * Nx])

    project = dg.create.projection(g, g_fine)
    vec = dg.evaluate(lambda y, x: np.sin(x) * np.sin(y), g_fine)
    project = project.dot(vec)
    projectE = dg.evaluate(lambda y, x: np.sin(x) * np.sin(y), g)
    error = np.sum((project - projectE) ** 2) / np.sum(project ** 2)
    print(f"Error is {np.sqrt(error)} (should be small)!")
    assert np.isclose(np.sqrt(error), 0, atol=1e-5)



import scipy.sparse as sparse

from .. import enums
from .derivative import dx, jump


def elliptic(grid, bcs, directions, sigma, jumpfactor=1):
    left = [
        dx(i, grid, enums.inverse_bc(bcs[i]), enums.inverse_dir(directions[i]))
        for i in range(0, grid.ndim)
    ]
    right = [dx(i, grid, bcs[i], directions[i]) for i in range(0, grid.ndim)]
    jumps = [jump(i, grid, bcs[i]) for i in range(0, grid.ndim)]
    matrix = -left[0] @ sparse.diags(sigma) @ right[0] + jumpfactor * jumps[0]
    for i in range(1, grid.ndim):
        matrix += -left[i] @ sparse.diags(sigma) @ right[i] + jumpfactor * jumps[i]
    return matrix



import numpy as np


class Grid:
    def __init__(self, x0, x1, n, N):
        """Create from lists of same length, which is the dimension of the grid

        For the one-dimensional case numbers can be given which will be
        converted to lists
        Dimension is given by length of parameters
        """
        self.x0 = np.asarray(x0)
        try:
            len(x0)
        except TypeError:
            self.x0 = np.asarray([x0])

        self.x1 = np.asarray(x1)  # must be numpy array of length dimension
        try:
            len(x1)
        except TypeError:
            self.x1 = np.asarray([x1])
        self.n = np.asarray(n)  # must be numpy array of length dimension
        try:
            len(n)
        except TypeError:
            self.n = np.asarray([n])
        self.N = np.asarray(N)  # must be numpy array of length dimension
        try:
            len(N)
        except TypeError:
            self.N = np.asarray([N])
        # lengths must be consistent
        assert(len(self.x0) == len(self.x1))
        assert(len(self.x0) == len(self.n))
        assert(len(self.x0) == len(self.N))

    @property
    def n(self):
        return self.__n

    @property
    def N(self):
        return self.__N

    @property
    def x0(self):
        return self.__x0

    @property
    def x1(self):
        return self.__x1

    @property
    def ndim(self):
        """Return the number of dimensions in the grid"""
        return len(self.x0)

    @property
    def shape(self):
        """Shape of the grid : n*N"""
        return self.n * self.N

    @n.setter
    def n(self, n):
        self.__n = np.asarray(n)

    @N.setter
    def N(self, N):
        self.__N = np.asarray(N)

    @x0.setter
    def x0(self, x0):
        self.__x0 = np.asarray(x0)

    @x1.setter
    def x1(self, x1):
        self.__x1 = np.asarray(x1)

    def lx(self):
        return self.x1 - self.x0

    def h(self):
        return (self.x1 - self.x0) / self.N

    def size(self):
        """The total number of points in the grid: np.prod(grid.shape)"""
        return np.prod(self.shape)



import numpy as np


def delta(n):
    op = np.zeros((n, n))
    for i in range(0, n):
        op[i, i] = 1.0
    return op


def pipj(n):
    op = np.zeros((n, n))
    for i in range(0, n):
        op[i, i] = 2.0 / (2 * i + 1.0)
    return op


def pipj_inv(n):
    op = np.zeros((n, n))
    for i in range(0, n):
        op[i, i] = (2 * i + 1.0) / 2.0
    return op


def pidxpj(n):
    op = np.zeros((n, n))
    for i in range(0, n):
        for j in range(0, n):
            if (i < j) and ((i + j) % 2 != 0):
                op[i, j] = 2
    return op


def rirj(n):
    op = np.zeros((n, n))
    op.fill(1)
    return op


def rilj(n):
    op = np.zeros((n, n))
    op.fill(-1)
    for i in range(0, n):
        for j in range(0, n):
            if j % 2 == 0:
                op[i, j] = 1.0
    return op


def lirj(n):
    op = rilj(n)
    return np.transpose(op)


def lilj(n):
    op = np.zeros((n, n))
    op.fill(-1)
    for i in range(0, n):
        for j in range(0, n):
            if ((i + j) % 2) == 0:
                op[i, j] = 1.0
    return op


def ninj(n):
    op = np.zeros((n, n))
    for i in range(0, n):
        for j in range(0, n):
            if i == j + 1:
                op[i, j] = 2.0 / (2 * i + 1) / (2 * j + 1)
            if i == j - 1:
                op[i, j] = -2.0 / (2 * i + 1) / (2 * j + 1)
    op[0, 0] = 2
    return op


def backward(n):
    (x, w) = np.polynomial.legendre.leggauss(n)
    vander = np.polynomial.legendre.legvander(x, n - 1)
    return vander


def forward(n):
    vander = backward(n)
    return np.linalg.inv(vander)



# docstring displayed by help(pyfeltor)
""" The python version of the dg library, creation
"""
from .derivative import dx as dx
from .derivative import jump as jump
from .elliptic import elliptic as elliptic
from .interpolation import interpolation as interpolation
from .interpolation import projection as projection
from .weights import abscissas as abscissas
from .weights import grid_from_abscissas as grid_from_abscissas
from .weights import weights as weights



import numpy as np

from pyfeltor import dg

# Run with pytest-3 -s . to see stdout output


def function2d(y, x):
    rho = 0.20943951023931953  # pi/15
    delta = 0.050000000000000003
    # return np.where(

    if y <= np.pi:
        return delta * np.cos(x) - 1.0 / rho / np.cosh(
            (y - np.pi / 2.0) / rho
        ) / np.cosh((y - np.pi / 2.0) / rho)
    else:
        return delta * np.cos(x) + 1.0 / rho / np.cosh(
            (3.0 * np.pi / 2.0 - y) / rho
        ) / np.cosh((3.0 * np.pi / 2.0 - y) / rho)


def test_evaluation():
    n, Nx, Ny, Nz = 3, 9, 5, 4
    equi = dg.Grid([-7, -5, -10], [-3, 5, 10], [1, 1, 1], [n * Nz, n * Ny, n * Nx])
    z = dg.evaluate(lambda z, y, x: z * x, equi)
    y = dg.evaluate(lambda z, y, x: y, equi)
    x = dg.evaluate(lambda z, y, x: x, equi)
    print("Length of arrays", len(z), len(y), len(x))
    assert len(z) == equi.size()
    assert len(y) == equi.size()
    assert len(x) == equi.size()


def test_integration():
    n = 3
    Nx = 12
    Ny = 28
    g1d = dg.Grid(x0=1, x1=2, n=n, N=Nx)
    g2d = dg.Grid((0, 0), (2 * np.pi, 2 * np.pi), (n, n), (Ny, Nx))
    w1d = dg.create.weights(g1d)
    w2d = dg.create.weights(g2d)
    func1d = dg.evaluate(lambda x: np.exp(x), g1d)
    func2d = dg.evaluate(function2d, g2d)

    sol1d = np.exp(2) - np.exp(1)
    num1d = np.sum(w1d * func1d)
    print(f"Correct integral is {sol1d} while numerical is {num1d}")
    assert np.abs(sol1d - num1d) / sol1d < 1e-10

    sol2d = 0.0
    num2d = np.sum(w2d * func2d)
    print(f"Correct integral is {sol2d} while numerical is {num2d}")
    assert np.abs(sol2d - num2d) < 1e-10


def test_integrate():
    print("TEST OF dg.integrate")
    g1d = dg.Grid([1], [2], [3], [12])
    w1d = dg.create.weights(g1d)
    integral_num = dg.integrate(
        dg.evaluate(lambda x: np.cos(x), g1d), g1d, dg.direction.forward
    )
    integral_ana = dg.evaluate(lambda x: np.sin(x), g1d) - np.sin(g1d.x0[0])
    error = integral_ana - integral_num
    norm = np.sum(w1d * error ** 2)
    print(" Error norm of  1d integral function (forward) ", norm)
    assert np.isclose(norm, 0)
    integral_num = dg.integrate(
        dg.evaluate(lambda x: np.cos(x), g1d), g1d, dg.direction.backward
    )
    integral_ana = dg.evaluate(lambda x: np.sin(x), g1d) - np.sin(g1d.x1[0])
    error = integral_ana - integral_num
    norm = np.sum(w1d * error ** 2)
    print(" Error norm of  1d integral function (backward) ", norm)
    assert np.isclose(norm, 0)
